/* V1.00 Object and list handeling routines for C */
/* (c) 1992 James McKenzie */




#include "project.h"

void 
Obj_DefaultDestructor(Object *object)
{
  if (object == NULL) {
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Invalid object 		Op: destruct\n");
    abort();
  }
  free(object);
  return;

}

void 
Obj_DefaultCreator(Object *object)
{
}

void 
Obj_CreateList(List *list)
{

  list->head = NULL;
  list->tail = NULL;

}


void 
Obj_Init(Object *object)
{
  object->next = NULL;
  object->prev = NULL;
  object->destructor = Obj_DefaultDestructor;
}



void 
Obj_Delete(Object *object)
{

  if (object == NULL) {
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Invalid object 		Op: delete\n");
    abort();
  }
  if (((object->next) != NULL) || ((object->prev) != NULL)) {
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Object is still in list	Op: delete\n");
    abort();
  }
  (*object->destructor) (object);


}



void 
Obj_Remove(List * list, Object * object)
{

  if (object == NULL) {
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Invalid object 		Op: remove\n");
    abort();
  }
  if (object->prev == NULL) {
    list->head = (Object *) object->next;
  } else {
    ((Object *) object->prev)->next = object->next;
  }

  if (object->next == NULL) {
    list->tail = (Object *) object->prev;
  } else {
    ((Object *) object->next)->prev = object->prev;
  }

  object->prev = NULL;
  object->next = NULL;

}


void 
Obj_Insert(List *list, Object *object, int action)
{

  if (object == NULL) {
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Invalid object 		Op: insert\n");
    abort();
  }
  if (((object->next) != NULL) || ((object->prev) != NULL)) {
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Object allready in list	Op: insert\n");
    abort();
  }
  switch (action) {
  case ObjInsertHead:
    if (list->head == NULL) {
      object->prev = NULL;
      object->next = (caddr_t) list->head;
      list->head = object;
    } else {
      object->prev = NULL;
      object->next = (caddr_t) list->head;
      list->head->prev = (caddr_t) object;
      list->head = object;
    }
    if (list->tail == NULL) {
      list->tail = object;
    }
    break;
  case ObjInsertTail:
    if (list->tail == NULL) {
      object->next = NULL;
      object->prev = (caddr_t) list->tail;
      list->tail = object;
    } else {
      object->next = NULL;
      object->prev = (caddr_t) list->tail;
      list->tail->next = (caddr_t) object;
      list->tail = object;
    }
    if (list->head == NULL) {
      list->head = object;
    }
    break;
  default:
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Invalid action 		Op: insert\n");
    abort();
  }
}


void Obj_InsertAfter(List *list, Object *p_object, Object *object)
{

  if (object == NULL) {
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Invalid object 		Op: insertafter\n");
    abort();
  }
  if (p_object == NULL) {
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Invalid p_object 		Op: insertafter\n");
    abort();
  }
  if (((object->next) != NULL) || ((object->prev) != NULL)) {
    fprintf(stderr, "Object package\n");
    fprintf(stderr, "Error: Object allready in list	Op: insertafter\n");
    abort();
  }
  object->next = p_object->next;
  p_object->next = (caddr_t) object;
  object->prev = (caddr_t) p_object;


  if (object->next == NULL) {
    if (list->tail != p_object) {
      fprintf(stderr, "Object package\n");
      fprintf(stderr, "Error: Object p_object not in list	Op: insertafter\n");
      abort();

    }
    list->tail = object;
  }
}






void Obj_DeleteList(List *list)
{
  Object         *object;

  while ((object = list->head) != NULL) {
    Obj_Remove(list, object);
    Obj_Delete(object);
  }


}
